# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
import shutil

from tornado.httpclient import HTTPClientError
from tornado.web import HTTPError

from grader_labextension.registry import register_handler
from grader_labextension.handlers.base_handler import ExtensionBaseHandler, cache
from grader_labextension.services.request import RequestService
import tornado
import os


@register_handler(path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/?")
class AssignmentBaseHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments.
    """

    @cache(max_age=30)
    async def get(self, lecture_id: int):
        """Sends a GET request to the grader service and returns assignments of the lecture

        :param lecture_id: id of the lecture
        :type lecture_id: int
        """
        try:
            response = await self.request_service.request(
                method="GET",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments",
                header=self.grader_authentication_header,
                response_callback=self.set_service_headers
            )

            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        # Create directories for every assignment
        try:
            dirs = set(
                filter(lambda e: e[0] != ".", os.listdir(os.path.expanduser(f'{self.root_dir}/{lecture["code"]}'))))
            for assignment in response:
                if assignment["id"] not in dirs:
                    self.log.info(f'Creating directory {self.root_dir}/{lecture["code"]}/{assignment["id"]}')
                    os.makedirs(
                        os.path.expanduser(f'{self.root_dir}/{lecture["code"]}/{assignment["id"]}'), exist_ok=True
                    )
                try:
                    dirs.remove(assignment["id"])
                except KeyError:
                    pass
        except FileNotFoundError:
            pass

        self.write(json.dumps(response))

    async def post(self, lecture_id: int):
        """Sends post-request to the grader service to create an assignment

        :param lecture_id: id of the lecture in which the new assignment is
        :type lecture_id: int
        """

        data = tornado.escape.json_decode(self.request.body)
        try:
            response = await self.request_service.request(
                method="POST",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments",
                body=data,
                header=self.grader_authentication_header,
            )

            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)
        # if we did not get an error when creating the assignment (i.e. the user is authorized etc.) then we can
        # create the directory structure if it does not exist yet
        os.makedirs(
            os.path.expanduser(f'{self.root_dir}/{lecture["code"]}/{response["id"]}'),
            exist_ok=True
        )
        os.makedirs(
            os.path.expanduser(f"{self.root_dir}/source/{lecture['code']}/{response['id']}"),
            exist_ok=True,
        )
        os.makedirs(
            os.path.expanduser(f"{self.root_dir}/release/{lecture['code']}/{response['id']}"),
            exist_ok=True,
        )
        self.write(json.dumps(response))


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/?"
)
class AssignmentObjectHandler(ExtensionBaseHandler):
    """
        Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}.
        """

    async def put(self, lecture_id: int, assignment_id: int):
        """Sends a PUT-request to the grader service to update a assignment

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """

        data = tornado.escape.json_decode(self.request.body)
        try:
            response = await self.request_service.request(
                method="PUT",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                body=data,
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)
        self.write(json.dumps(response))

    @cache(max_age=30)
    async def get(self, lecture_id: int, assignment_id: int):
        """Sends a GET-request to the grader service to get a specific assignment

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the specific assignment
        :type assignment_id: int
        """

        query_params = RequestService.get_query_string(
            {
                "instructor-version": self.get_argument("instructor-version", None),
            }
        )

        try:
            response = await self.request_service.request(
                method="GET",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}{query_params}",
                header=self.grader_authentication_header,
                response_callback=self.set_service_headers
            )
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        os.makedirs(
            os.path.expanduser(f'{self.root_dir}/{lecture["code"]}/{response["id"]}'),
            exist_ok=True
        )
        self.write(json.dumps(response))

    async def delete(self, lecture_id: int, assignment_id: int):
        """Sends a DELETE-request to the grader service to "soft"-delete a assignment

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """

        try:
            assignment = await self.request_service.request(
                method="GET",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            await self.request_service.request(
                method="DELETE",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
                decode_response=False
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        self.log.warn(f'Deleting directory {self.root_dir}/{lecture["code"]}/{assignment["id"]}')
        shutil.rmtree(os.path.expanduser(f'{self.root_dir}/{lecture["code"]}/{assignment["id"]}'), ignore_errors=True)
        self.write("OK")


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/properties\/?"
)
class AssignmentPropertiesHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/properties.
    """

    async def get(self, lecture_id: int, assignment_id: int):
        """Sends a GET-request to the grader service and returns the properties of an assignment

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """

        try:
            response = await self.request_service.request(
                method="GET",
                endpoint=f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}/properties",
                header=self.grader_authentication_header,
                response_callback=self.set_service_headers
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)
        self.write(json.dumps(response))




# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import functools
import json
import traceback
from typing import Optional, Awaitable

from tornado import httputil
from tornado.web import HTTPError

from grader_labextension.api.models.error_message import ErrorMessage
from grader_labextension.services.request import RequestService
from jupyter_server.base.handlers import APIHandler
import os
from tornado.httpclient import HTTPClientError, HTTPResponse
from traitlets.config.configurable import SingletonConfigurable
from traitlets.traitlets import Unicode


def cache(max_age: int):
    if max_age < 0:
        raise ValueError("max_age must be larger than 0!")

    def wrapper(handler_method):
        @functools.wraps(handler_method)
        async def request_handler_wrapper(self: "ExtensionBaseHandler", *args, **kwargs):
            self.set_header("Cache-Control", f"max-age={max_age}, must-revalidate, private")
            return await handler_method(self, *args, **kwargs)
        return request_handler_wrapper
    return wrapper


class HandlerConfig(SingletonConfigurable):
    hub_api_url = Unicode(os.environ.get("JUPYTERHUB_API_URL"), help="The url of the hubs api.").tag(config=True)
    hub_api_token = Unicode(os.environ.get("JUPYTERHUB_API_TOKEN"), help="The authorization token to access the hub api").tag(config=True)
    hub_user = Unicode(os.environ.get("JUPYTERHUB_USER"), help="The user name in jupyter hub.").tag(config=True)
    service_base_url = Unicode(
        os.environ.get("GRADER_BASE_URL", "/services/grader"),
        help="Base URL to use for each request to the grader service",
    ).tag(config=True)


class ExtensionBaseHandler(APIHandler):
    """
    BaseHandler for all server-extension handler
    """

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def set_service_headers(self, response: HTTPResponse):
        for header in response.headers.get_list("Cache-Control"):
            self.set_header("Cache-Control", header)

    request_service = RequestService.instance()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.root_dir = os.path.expanduser(self.settings["server_root_dir"])

    @property
    def service_base_url(self):
        return HandlerConfig.instance().service_base_url

    @property
    def grader_authentication_header(self):
        """Returns the authentication header

        :return: authentication header
        :rtype: dict
        """

        return dict(Authorization="Token " + HandlerConfig.instance().hub_api_token)

    @property
    def user_name(self):
        return self.current_user["name"]

    async def get_lecture(self, lecture_id) -> dict:
        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            return lecture
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

    async def get_assignment(self, lecture_id, assignment_id):
        try:
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
            return assignment
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

    def write_error(self, status_code: int, **kwargs) -> None:
        self.set_header('Content-Type', 'application/json')
        _, e, _ = kwargs.get("exc_info", (None, None, None))
        error = httputil.responses.get(status_code, "Unknown")
        reason = None
        if e and isinstance(e, HTTPError) and e.reason:
            reason = e.reason
        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            # in debug mode, try to send a traceback
            lines = []
            for line in traceback.format_exception(*kwargs["exc_info"]):
                lines.append(line)
            self.finish(json.dumps(
                ErrorMessage(status_code, error, self.request.path, reason, traceback=json.dumps(lines)).to_dict()))
        else:
            self.finish(json.dumps(ErrorMessage(status_code, error, self.request.path, reason).to_dict()))



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
import logging
import subprocess
from typing import List, Dict, Union, Tuple
from urllib.parse import urlparse, ParseResultBytes

from traitlets.config.configurable import Configurable
from traitlets.config.loader import Config
from traitlets.traitlets import Int, TraitError, Unicode, validate
import os
import posixpath
import shlex
from datetime import datetime
import shutil
import sys


class GitError(Exception):
    def __init__(self, error: str):
        self.error = error

    def __str__(self):
        return self.error

    def __repr__(self) -> str:
        return self.__str__()


class RemoteStatus(enum.Enum):
    up_to_date = 1
    pull_needed = 2
    push_needed = 3
    divergent = 4


class GitService(Configurable):
    git_access_token = Unicode(os.environ.get("JUPYTERHUB_API_TOKEN"), allow_none=False).tag(config=True)
    git_service_url = Unicode(
        f'{os.environ.get("GRADER_HOST_URL", "http://127.0.0.1:4010")}{os.environ.get("GRADER_GIT_BASE_URL", "/services/grader/git")}',
        allow_none=False).tag(config=True)

    def __init__(self, server_root_dir: str, lecture_code: str, assignment_id: int, repo_type: str,
                 force_user_repo=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger(str(self.__class__))
        self._git_version = None
        self.git_root_dir = server_root_dir
        self.lecture_code = lecture_code
        self.assignment_id = assignment_id
        self.repo_type = repo_type
        if self.repo_type == "assignment" or force_user_repo:
            self.path = os.path.join(self.git_root_dir, self.lecture_code, str(self.assignment_id))
        else:
            self.path = os.path.join(self.git_root_dir, self.repo_type, self.lecture_code, str(self.assignment_id))
        self.log.info(f"New git service working in {self.path}")
        os.makedirs(self.path, exist_ok=True)

        self.log.info("git_access_token: " + self.git_access_token)
        url_parsed = urlparse(self.git_service_url)
        self.log.info(f"git_service_url: " + self.git_service_url)
        self.git_http_scheme: str = url_parsed.scheme
        self.git_remote_url: str = url_parsed.netloc + url_parsed.path
        self.log.info("git_http_scheme: " + self.git_http_scheme)
        self.log.info("git_remote_url: " + self.git_remote_url)

    def push(self, origin: str, force=False):
        """Pushes commits on the remote

        Args:
            origin (str): the remote
            force (bool, optional): states if the operation should be forced. Defaults to False.
        """
        self.log.info(f"Pushing remote {origin} for {self.path}")
        self._run_command(f"git push {origin} main" + (" --force" if force else ""), cwd=self.path)

    def set_remote(self, origin: str, sub_id=None):
        """Set a remote in the local repository

        Args:
            origin (str): the remote
            sub_id ([type], optional): a query param for the feedback pull. Defaults to None.
        """
        self.log.info(f"Setting remote {origin} for {self.path}")
        url = posixpath.join(self.git_remote_url, self.lecture_code, str(self.assignment_id), self.repo_type)
        try:
            if sub_id is None:
                self._run_command(
                    f"git remote add {origin} {self.git_http_scheme}://oauth:{self.git_access_token}@{url}",
                    cwd=self.path)
            else:
                self.log.info(f"Setting remote with sub_id {sub_id}")
                self._run_command(
                    f"git remote add {origin} {self.git_http_scheme}://oauth:{self.git_access_token}@{posixpath.join(url, sub_id)}",
                    cwd=self.path)

        except GitError as e:
            self.log.error("GitError:\n" + e.error)
            self.log.info(f"Remote set: Updating remote {origin} for {self.path}")
            if sub_id is None:
                self._run_command(
                    f"git remote set-url {origin} {self.git_http_scheme}://oauth:{self.git_access_token}@{url}",
                    cwd=self.path)
            else:
                self.log.info(f"Setting remote with sub_id {sub_id}")
                self._run_command(
                    f"git remote set-url {origin} {self.git_http_scheme}://oauth:{self.git_access_token}@{posixpath.join(url, sub_id)}",
                    cwd=self.path)

    def delete_remote(self, origin: str):
        raise NotImplementedError()

    def switch_branch(self, branch: str):
        """Switches into another branch

        Args:
            branch (str): the branch name
        """
        self.log.info(f"Fetching all at path {self.path}")
        self._run_command(f"git fetch --all", cwd=self.path)
        self.log.info(f"Switching to branch {branch} at path {self.path}")
        self._run_command(f"git checkout {branch}", cwd=self.path)

    def fetch_all(self):
        self.log.info(f"Fetching all at path {self.path}")
        self._run_command(f"git fetch --all", cwd=self.path)

    def pull(self, origin: str, branch="main", force=False):
        """Pulls a repository

        Args:
            origin (str): the remote
            branch (str, optional): the branch name. Defaults to "main".
            force (bool, optional): states if the operation should be forced. Defaults to False.
        """
        if force:
            self.log.info(f"Pulling remote {origin}")
            out = self._run_command(
                f'sh -c "git clean -fd && git fetch {origin} && git reset --hard {origin}/{branch}"', cwd=self.path,
                capture_output=True)
            self.log.info(out)
            # self._run_command(f'sh -c "git fetch --all && git reset --mixed {origin}/main"',cwd=self.path)
        else:
            self._run_command(f"git pull {origin} {branch}", cwd=self.path)

    def init(self, force=False):
        """Initiates a local repository

        Args:
            force (bool, optional): states if the operation should be forced. Defaults to False.
        """
        if not self.is_git() or force:
            self.log.info(f"Calling init for {self.path}")
            if self.git_version < (2, 28):
                self._run_command(f"git init", cwd=self.path)
                self._run_command("git checkout -b main", cwd=self.path)
            else:
                self._run_command(f"git init -b main", cwd=self.path)

    def is_git(self):
        """Checks if the directory is a local repository

        Returns:
            bool: states if the directory is a repository
        """
        return os.path.exists(os.path.join(self.path, ".git"))

    def commit(self, m=str(datetime.now())):
        """Commits the staged changes

        Args:
            m (str, optional): the commit message. Defaults to str(datetime.now()).
        """
        # self.log.info("Adding all files")
        # self._run_command(f'git add -A', cwd=self.path)
        # self.log.info("Committing repository")
        # self._run_command(f'git commit -m "{m}"', cwd=self.path)
        self.log.info(f"Adding all files and committing in {self.path}")
        self._run_command(f'sh -c \'git add -A && git commit --allow-empty -m "{m}"\'', cwd=self.path)

    def set_author(self, author):
        # TODO: maybe ask user to specify their own choices
        self._run_command(f'git config user.name "{author}"', cwd=self.path)
        self._run_command(f'git config user.email "sample@mail.com"', cwd=self.path)

    def clone(self, origin: str, force=False):
        """Clones the repository

        Args:
            origin (str): the remote
            force (bool, optional): states if the operation should be forced. Defaults to False.
        """
        self.init(force=force)
        self.set_remote(origin=origin)
        self.pull(origin=origin, force=force)

    def delete_repo_contents(self, include_git=False):
        """Deletes the contents of the git service

        Args:
            include_git (bool, optional): states if the .git directory should also be deleted. Defaults to False.
        """
        for root, dirs, files in os.walk(self.path):
            for f in files:
                os.unlink(os.path.join(root, f))
                self.log.info(f"Deleted {os.path.join(root, f)} from {self.git_root_dir}")
            for d in dirs:
                if d != ".git" or include_git:
                    shutil.rmtree(os.path.join(root, d))
                    self.log.info(f"Deleted {os.path.join(root, d)} from {self.git_root_dir}")

    # Note: dirs_exist_ok was only added in Python 3.8
    def copy_repo_contents(self, src: str):
        """copies repo contents from src to the git path

        Args:
            src (str): path where the to be copied files reside
        """
        self.log.info(f"Copying repository contents from {src} to {self.path}")
        ignore = shutil.ignore_patterns(".git", "__pycache__")
        if sys.version_info.major == 3 and sys.version_info.minor >= 8:
            shutil.copytree(src, self.path, ignore=ignore, dirs_exist_ok=True)
        else:
            for item in os.listdir(src):
                s = os.path.join(src, item)
                d = os.path.join(self.path, item)
                if os.path.isdir(s):
                    shutil.copytree(s, d, ignore=ignore)
                else:
                    shutil.copy2(s, d)

    def check_remote_status(self, origin: str, branch: str) -> RemoteStatus:
        untracked, added, modified, deleted = self.git_status(hidden_files=False)
        local_changes = len(untracked) > 0 or len(added) > 0 or len(modified) > 0 or len(deleted) > 0
        if self.local_branch_exists(branch):
            local = self._run_command(f"git rev-parse {branch}", cwd=self.path, capture_output=True).strip()
        else:
            local = None
        if self.remote_branch_exists(origin, branch):
            remote = self._run_command(f"git rev-parse {origin}/{branch}", cwd=self.path, capture_output=True).strip()
        else:
            if len(untracked) + len(added) + len(modified) == 0:
                return RemoteStatus.up_to_date  # if we don't have remote and no files we are up-to-date
            return RemoteStatus.push_needed

        if local is None and remote:
            if local_changes:
                return RemoteStatus.divergent
            return RemoteStatus.pull_needed

        if local == remote:
            if local_changes:
                return RemoteStatus.push_needed
            return RemoteStatus.up_to_date

        base = self._run_command(f"git merge-base {branch} {origin}/{branch}", cwd=self.path, capture_output=True).strip()

        if local == base:
            return RemoteStatus.pull_needed
        elif remote == base:
            return RemoteStatus.push_needed
        else:
            return RemoteStatus.divergent

    def git_status(self, hidden_files: bool = False) -> Tuple[List[str], List[str], List[str], List[str]]:
        files = self._run_command("git status --porcelain", cwd=self.path, capture_output=True)
        untracked, added, modified, deleted = [], [], [], []
        for line in files.splitlines():
            k, v = line.split(maxsplit=1)
            if v[0] == "." and not hidden_files:  # TODO: implement nested check for hidden files
                continue
            if k == "??":
                untracked.append(v)
            elif k == "A":
                added.append(v)
            elif k == "M":
                modified.append(v)
            elif k == "D":
                deleted.append(v)
        return untracked, added, modified, deleted

    def local_branch_exists(self, branch: str) -> bool:
        ret_code = self._run_command(f"git rev-parse --quiet --verify {branch}", cwd=self.path, check=False).returncode
        if ret_code == 0:
            return True
        else:
            return False

    def remote_branch_exists(self, origin: str, branch: str) -> bool:
        ret_code = self._run_command(f"git ls-remote --exit-code {origin}  {branch}", cwd=self.path,
                                     check=False).returncode
        if ret_code == 0:
            return True
        else:
            return False

    def get_log(self, history_count=10) -> List[Dict[str, str]]:
        """
        Execute git log command & return the result.
        """
        cmd = f'git log --pretty=format:%H%n%an%n%at%n%D%n%s -{history_count}'
        my_output = self._run_command(cmd, cwd=self.path, capture_output=True)

        result = []
        line_array = my_output.splitlines()
        previous_commit_offset = 5
        self.log.info(f"Found {len(line_array) // previous_commit_offset} commits in history")

        for i in range(0, len(line_array), previous_commit_offset):
            commit = {
                "commit": line_array[i],
                "author": line_array[i + 1],
                "date": datetime.utcfromtimestamp(int(line_array[i + 2])).isoformat("T", "milliseconds") + "Z",
                # "date": line_array[i + 2],
                "ref": line_array[i + 3],
                "commit_msg": line_array[i + 4],
                "pre_commit": "",
            }

            if i + previous_commit_offset < len(line_array):
                commit["pre_commit"] = line_array[i + previous_commit_offset]

            result.append(commit)

        return result

    @property
    def git_version(self):
        """Return the git version

        Returns:
            tuple: the git version
        """
        if self._git_version is None:
            try:
                version = self._run_command("git --version", capture_output=True)
            except GitError:
                return tuple()
            version = version.split(" ")[2]
            self._git_version = tuple([int(v) for v in version.split(".")])
        return self._git_version

    def _run_command(self, command, cwd=None, capture_output=False, check=True) -> Union[str, subprocess.CompletedProcess]:
        """Starts a sub process and runs a cmd command

        Args:
            command str: command that is getting run.
            cwd (str, optional): states where the command is getting run. Defaults to None.
            capture_output (bool, optional): states if output is getting saved. Defaults to False.
            check (bool, optional): whether to raise a GitError if process fails.
        Raises:
            GitError: returns appropriate git error 

        Returns:
            str: command output
        """
        ret = None
        try:
            self.log.info(f"Running: {command}")
            ret = subprocess.run(shlex.split(command), cwd=cwd, capture_output=True, text=True)
            ret.check_returncode()
            if capture_output:
                return ret.stdout
            else:
                return ret
        except subprocess.CalledProcessError as e:
            raise GitError(ret.stderr.replace("\n", ""))
        except FileNotFoundError as e:
            raise GitError(e.strerror)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
import shutil
from http import HTTPStatus
from urllib.parse import unquote, quote
from tornado.web import HTTPError

from grader_convert.converters.base import GraderConvertException
from grader_convert.converters.generate_assignment import GenerateAssignment
from .base_handler import ExtensionBaseHandler, cache
from ..api.models.submission import Submission
from ..registry import register_handler
from ..services.git import GitError, GitService
from tornado.httpclient import HTTPClientError, HTTPResponse


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/generate\/?"
)
class GenerateHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/generate.
    """

    async def put(self, lecture_id: int, assignment_id: int):
        """Generates the release files from the source files of a assignment

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        """
        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)
        code = lecture["code"]
        a_id = assignment["id"]

        output_dir = f"{self.root_dir}/release/{code}/{a_id}"
        os.makedirs(
            os.path.expanduser(output_dir),
            exist_ok=True,
        )

        generator = GenerateAssignment(
            input_dir=f"{self.root_dir}/source/{code}/{a_id}",
            output_dir=output_dir,
            file_pattern="*.ipynb",
            copy_files=True  # Always copy files from source to release
        )
        generator.force = True

        try:
            # delete contents of output directory since we might have chosen to disallow files
            self.log.info("Deleting files in release directory")
            shutil.rmtree(output_dir)
            os.mkdir(output_dir)
        except Exception as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=str(e))

        self.log.info("Starting GenerateAssignment converter")
        try:
            generator.start()
        except Exception as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.CONFLICT, reason=str(e))
        try:
            gradebook_path = os.path.join(generator._output_directory, "gradebook.json")
            os.remove(gradebook_path)
            self.log.info(f"Successfully deleted {gradebook_path}")
        except OSError as e:
            self.log.error(f"Could delete {gradebook_path}! Error: {e.strerror}")
        self.log.info("GenerateAssignment conversion done")
        self.write("OK")


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/remote-status\/(?P<repo>\w*)\/?"
)
class GitRemoteStatusHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/remote_status/{repo}.
    """

    @cache(max_age=15)
    async def get(self, lecture_id: int, assignment_id: int, repo: str):
        if repo not in {"assignment", "source", "release"}:
            self.log.error(HTTPStatus.NOT_FOUND)
            raise HTTPError(HTTPStatus.NOT_FOUND, reason=f"Repository {repo} does not exist")
        lecture = await self.get_lecture(lecture_id)
        assignment = await self.get_assignment(lecture_id, assignment_id)
        git_service = GitService(
            server_root_dir=self.root_dir,
            lecture_code=lecture["code"],
            assignment_id=assignment["id"],
            repo_type=repo,
            config=self.config,
            force_user_repo=True if repo == "release" else False,
        )
        try:
            if not git_service.is_git():
                git_service.init()
                git_service.set_author(author=self.user_name)
            git_service.set_remote(f"grader_{repo}")
            git_service.fetch_all()
            status = git_service.check_remote_status(f"grader_{repo}", "main")
        except GitError as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e)
        self.write(status.name)


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/log\/(?P<repo>\w*)\/?"
)
class GitLogHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/log/{repo}.
    """

    @cache(max_age=15)
    async def get(self, lecture_id: int, assignment_id: int, repo: str):
        """
        Sends a GET request to the grader service to get the logs of a given repo.

        :param lecture_id: id of the lecture
        :param assignment_id: id of the assignment
        :param repo: repo name
        :return: logs of git repo
        """
        if repo not in {"assignment", "source", "release"}:
            self.log.error(HTTPStatus.NOT_FOUND)
            raise HTTPError(HTTPStatus.NOT_FOUND, reason=f"Repository {repo} does not exist")
        n_history = int(self.get_argument("n", "10"))
        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        git_service = GitService(
            server_root_dir=self.root_dir,
            lecture_code=lecture["code"],
            assignment_id=assignment["id"],
            repo_type=repo,
            config=self.config,
            force_user_repo=True if repo == "release" else False,
        )
        try:
            if not git_service.is_git():
                git_service.init()
                git_service.set_author(author=self.user_name)
            git_service.set_remote(f"grader_{repo}")
            git_service.fetch_all()
            if git_service.local_branch_exists("main"):  # at least main should exist
                logs = git_service.get_log(n_history)
            else:
                logs = []
        except GitError as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e)

        self.write(json.dumps(logs))


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/pull\/(?P<repo>\w*)\/?"
)
class PullHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/pull/{repo}.
    """

    async def get(self, lecture_id: int, assignment_id: int, repo: str):
        """Creates a local repository and pulls the specified repo type

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param repo: type of the repository
        :type repo: str
        """
        if repo not in {"assignment", "source", "release"}:
            self.log.error(HTTPStatus.NOT_FOUND)
            raise HTTPError(HTTPStatus.NOT_FOUND, reason=f"Repository {repo} does not exist")
        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        git_service = GitService(
            server_root_dir=self.root_dir,
            lecture_code=lecture["code"],
            assignment_id=assignment["id"],
            repo_type=repo,
            config=self.config,
            force_user_repo=repo == "release",
        )
        try:
            if not git_service.is_git():
                git_service.init()
                git_service.set_author(author=self.user_name)
            git_service.set_remote(f"grader_{repo}")
            git_service.pull(f"grader_{repo}", force=True)
            self.write("OK")
        except GitError as e:
            self.log.error("GitError:\n" + e.error)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e.error)


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/push\/(?P<repo>\w*)\/?"
)
class PushHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/push/{repo}.
    """

    async def put(self, lecture_id: int, assignment_id: int, repo: str):
        """Pushes from the local repositories to remote
            If the repo type is release, it also generate the release files and updates the assignment properties in the grader service

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param repo: type of the repository
        :type repo: str
        """
        if repo not in {"assignment", "source", "release"}:
            self.write_error(404)
        commit_message = self.get_argument("commit-message", None)
        submit = self.get_argument("submit", "false") == "true"
        if repo == "source" and (commit_message is None or commit_message == ""):
            self.log.error("Commit message was not found")
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Commit message was not found")

        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)

        git_service = GitService(
            server_root_dir=self.root_dir,
            lecture_code=lecture["code"],
            assignment_id=assignment["id"],
            repo_type=repo,
            config=self.config,
        )

        if repo == "release":
            git_service.delete_repo_contents(include_git=True)
            src_path = GitService(
                self.root_dir,
                lecture["code"],
                assignment["id"],
                repo_type="source",
                config=self.config,
            ).path
            git_service.copy_repo_contents(src=src_path)

            # call nbconvert before pushing
            generator = GenerateAssignment(
                input_dir=src_path,
                output_dir=git_service.path,
                file_pattern="*.ipynb",
                copy_files=True  # Always copy files from source to release
            )
            generator.force = True

            try:
                # delete contents of output directory since we might have chosen to disallow files
                self.log.info("Deleting files in release directory")
                shutil.rmtree(git_service.path)
                os.mkdir(git_service.path)
            except Exception as e:
                self.log.error(e)
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=str(e))

            self.log.info("Starting GenerateAssignment converter")
            try:
                generator.start()
                self.log.info("GenerateAssignment conversion done")
            except GraderConvertException as e:
                self.log.error("Converting failed: Error converting notebook!", exc_info=True)

                raise HTTPError(409, reason=str(e))
            try:
                gradebook_path = os.path.join(git_service.path, "gradebook.json")
                self.log.info(f"Reading gradebook file: {gradebook_path}")
                with open(gradebook_path, "r") as f:
                    gradebook_json: dict = json.load(f)
            except FileNotFoundError:
                self.log.error(f"Cannot find gradebook file: {gradebook_path}")
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR,
                                reason=f"Cannot find gradebook file: {gradebook_path}")

            self.log.info(f"Setting properties of assignment from {gradebook_path}")
            response: HTTPResponse = await self.request_service.request(
                "PUT",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}/properties",
                header=self.grader_authentication_header,
                body=gradebook_json,
                decode_response=False,
            )
            if response.code == 200:
                self.log.info("Properties set for assignment")
            else:
                self.log.error(
                    f"Could not set assignment properties! Error code {response.code}"
                )

            try:
                os.remove(gradebook_path)
                self.log.info(f"Successfully deleted {gradebook_path}")
            except OSError as e:
                self.log.error(
                    f"Cannot delete {gradebook_path}! Error: {e.strerror}\nAborting push!"
                )
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR,
                                reason=f"Cannot delete {gradebook_path}! Error: {e.strerror}\nAborting push!")

        self.log.info(f"File contents of {repo}: {git_service.path}")
        self.log.info(",".join(os.listdir(git_service.path)))

        try:
            if not git_service.is_git():
                git_service.init()
                git_service.set_author(author=self.user_name)
                # TODO: create .gitignore file
            git_service.set_remote(f"grader_{repo}")
        except GitError as e:
            self.log.error("GitError:\n" + e.error)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e.error)

        try:
            git_service.commit(m=commit_message)
        except GitError as e:
            self.log.error("GitError:\n" + e.error)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e.error)

        try:
            git_service.push(f"grader_{repo}", force=True)
        except GitError as e:
            self.log.error("GitError:\n" + e.error)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=str(e.error))

        if submit and repo == "assignment":
            self.log.info(f"Submitting assignment {assignment_id}!")
            try:
                latest_commit_hash = git_service.get_log(history_count=1)[0]["commit"]
                submission = Submission(commit_hash=latest_commit_hash)
                response = await self.request_service.request(
                    "POST",
                    f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}/submissions",
                    body=submission.to_dict(),
                    header=self.grader_authentication_header,
                )
                self.write(json.dumps(response))
                return
            except (KeyError, IndexError) as e:
                self.log.error(e)
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e)
            except HTTPClientError as e:
                self.log.error(e.response)
                raise HTTPError(e.code, reason=e.response.reason)

        self.write("OK")


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/reset\/?"
)
class ResetHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/reset.
    """

    async def get(self, lecture_id: int, assignment_id: int):
        """
        Sends a GET request to the grader service that resets the user repo.

        :param lecture_id: id of the lecture
        :param assignment_id: id of the assignment
        :return: void
        """
        try:
            await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}/reset",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.log.error(e.response)
            raise HTTPError(e.code, reason=e.response.reason)
        self.write("OK")


@register_handler(
    path=r"\/(?P<lecture_id>\d*)\/(?P<assignment_id>\d*)\/(?P<notebook_name>.*)"
)
class NotebookAccessHandler(ExtensionBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/assignments/{assignment_id}/{notebook_name}.
    """

    async def get(self, lecture_id: int, assignment_id: int, notebook_name: str):
        """
        Sends a GET request to the grader service to access notebook and redirect to it.
        :param lecture_id: id of the lecture
        :param assignment_id: id of the assignment
        :param notebook_name: notebook name
        :return: void
        """
        notebook_name = unquote(notebook_name)

        try:
            lecture = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}",
                header=self.grader_authentication_header,
            )
            assignment = await self.request_service.request(
                "GET",
                f"{self.service_base_url}/lectures/{lecture_id}/assignments/{assignment_id}",
                header=self.grader_authentication_header,
            )
        except HTTPClientError as e:
            self.set_status(e.code)
            self.write_error(e.code)
            return

        git_service = GitService(
            server_root_dir=self.root_dir,
            lecture_code=lecture["code"],
            assignment_id=assignment["id"],
            repo_type="release",
            config=self.config,
            force_user_repo=True,
        )

        if not git_service.is_git():
            try:
                git_service.init()
                git_service.set_author(author=self.user_name)
                git_service.set_remote(f"grader_release")
                git_service.pull(f"grader_release", force=True)
                self.write("OK")
            except GitError as e:
                self.log.error("GitError:\n" + e.error)
                self.write_error(400)

        try:
            username = self.get_current_user()["name"]
        except TypeError as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason=e)

        url = f'/user/{username}/lab/tree/{lecture["code"]}/{assignment["id"]}/{quote(notebook_name)}'
        self.log.info(f"Redirecting to {url}")
        self.redirect(url)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_labextension.api.models.base_model_ import Model
from grader_labextension.api import util


class User(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, name=None):  # noqa: E501
        """User - a model defined in OpenAPI

        :param name: The name of this User.  # noqa: E501
        :type name: str
        """
        self.openapi_types = {
            'name': str
        }

        self.attribute_map = {
            'name': 'name'
        }

        self._name = name

    @classmethod
    def from_dict(cls, dikt) -> 'User':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The User of this User.  # noqa: E501
        :rtype: User
        """
        return util.deserialize_model(dikt, cls)

    @property
    def name(self):
        """Gets the name of this User.


        :return: The name of this User.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this User.


        :param name: The name of this User.
        :type name: str
        """

        self._name = name



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

"""
grader_labextension setup
"""
import json
import sys
from pathlib import Path

import setuptools

HERE = Path(__file__).parent.resolve()

# Get the package info from package.json
pkg_json = json.loads((HERE / "package.json").read_bytes())

# The name of the project
name = "grader_labextension"

lab_path = (HERE / pkg_json["jupyterlab"]["outputDir"])

# Representative files that should exist after a successful build
ensured_targets = [
    str(lab_path / "package.json"),
    str(lab_path / "static/style.js")
]

labext_name = pkg_json["name"]

data_files_spec = [
    ("share/jupyter/labextensions/%s" % labext_name, str(lab_path.relative_to(HERE)), "**"),
    ("share/jupyter/labextensions/%s" % labext_name, str("."), "install.json"),
    ("etc/jupyter/jupyter_server_config.d",
     "jupyter-config/server-config", "grader_labextension.json"),
    # For backward compatibility with notebook server
    ("etc/jupyter/jupyter_notebook_config.d",
     "jupyter-config/nb-config", "grader_labextension.json"),
]

long_description = (HERE / "README.md").read_text()

version = (
    pkg_json["version"]
    .replace("-alpha.", "a")
    .replace("-beta.", "b")
    .replace("-rc.", "rc")
) 

setup_args = dict(
    name=name,
    version=version,
    url=pkg_json["homepage"],
    author=pkg_json["author"]["name"],
    author_email=pkg_json["author"]["email"],
    description=pkg_json["description"],
    license=pkg_json["license"],
    license_file="LICENSE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "jupyter_server>=1.6,<2",
        "urllib3>=1.26.6",
        "traitlets>=5.0.5",
        "grader-convert>=0.1.2",
    ],
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.7",
    platforms="Linux, Mac OS X, Windows",
    keywords=["Jupyter", "JupyterLab", "JupyterLab3"],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: Jupyter",
        "Framework :: Jupyter :: JupyterLab",
        "Framework :: Jupyter :: JupyterLab :: 3",
        "Framework :: Jupyter :: JupyterLab :: Extensions",
        "Framework :: Jupyter :: JupyterLab :: Extensions :: Prebuilt",
    ],
)

try:
    from jupyter_packaging import (
        wrap_installers,
        npm_builder,
        get_data_files
    )
    post_develop = npm_builder(
        build_cmd="install:extension", source_dir="src", build_dir=lab_path
    )
    setup_args["cmdclass"] = wrap_installers(post_develop=post_develop, ensured_targets=ensured_targets)
    setup_args["data_files"] = get_data_files(data_files_spec)
except ImportError as e:
    import logging
    logging.basicConfig(format="%(levelname)s: %(message)s")
    logging.warning("Build tool `jupyter-packaging` is missing. Install it with pip or conda.")
    if not ("--name" in sys.argv or "--version" in sys.argv):
        raise e

if __name__ == "__main__":
    setuptools.setup(**setup_args)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import shutil
import sys
from pathlib import Path
from grader_labextension.services.request import RequestService
from jupyter_server.serverapp import ServerApp, ServerWebApplication
from traitlets.config.loader import Config

from ._version import __version__

HERE = Path(__file__).parent.resolve()

with (HERE / "labextension" / "package.json").open() as fid:
    data = json.load(fid)


def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": data["name"]
    }]


# unused import to register handlers
from grader_labextension import handlers
from grader_labextension.registry import HandlerPathRegistry


def validate_system_environment():
    if sys.version_info.major < 3 or sys.version_info.minor < 7:
        raise RuntimeError("This extension needs Python version 3.7 or above to run!")
    if shutil.which("git") is None:
        raise RuntimeError("No git executable found! Git is necessary to run the extension!")


def setup_handlers(web_app: ServerWebApplication, config: Config):
    host_pattern = ".*$"
    # RequestService.config = web_app.config
    base_url = web_app.settings["base_url"]
    log = logging.getLogger()
    log.info("#######################################################################")
    log.info(f'{web_app.settings["server_root_dir"]=}')
    log.info("base_url: " + base_url)
    handlers = HandlerPathRegistry.handler_list(base_url=base_url + "grader_labextension")
    log.info([str(h[0]) for h in handlers])
    web_app.add_handlers(host_pattern, handlers)


def _jupyter_server_extension_points():
    return [{
        "module": "grader_labextension"
    }]


def _load_jupyter_server_extension(server_app: ServerApp):
    """Registers the API handler to receive HTTP requests from the frontend extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """

    setup_handlers(server_app.web_app, server_app.config)
    server_app.log.info("Registered grading extension at URL path /grader_labextension")
